/**
 * @file decoder.h
 * @brief Header of the decoding functionality
 * @author Lukas Hostettler
 * @version 1.0
 * @date 2015-05-20
 */

#ifndef DECODER_H
#define DECODER_H

#ifdef __cplusplus
extern "C" {
#endif

#include "grid.h"

///@brief Stores the certainty for x Grid and y Grid
typedef IntPoint RotationDecoder;

///@brief Resets the rotation decoder to be completely unsure of its state
/// @var rot    Pointer to a rotationRecoder structure
void rotationDecoderReset(RotationDecoder* rot);

///@brief adds or removes certainty of the rotation decoder
/// @var rot        Pointer to a rotationRecoder structure
/// @var rotProb1   number of votes in favor  - number of votes against having the right orientation in x
/// @var rotProb2   number of votes in favor  - number of votes against having the right orientation in y
void rotationDecoderUpdate(RotationDecoder* rot, int rotProb1, int rotProb2);

///@brief If rotationDecoder finds that the current rotation is false it changes the means and the origin of dotInfo accordingly
///@returns the number of 90 degree rotations made by the function
///@var rot     Pointer to a rotationRecoder structure
///@var means   Pointer to the mean grid directions (will be changed if the rotation decoder indicates a change)
///@var dotInfo Pointer to a DotInformation structure (will be changed if the rotation decoder indicates a change)
int rotationDecoderUpdateMeans(RotationDecoder* rot, IntPoint* means, DotInformation* dotInfo);

///@brief Decreases cerainty in its current rotation state
///@var rot     Pointer to a rotationRecoder structure
void rotationDecoderDiminish(RotationDecoder* rot);

/// @brief  Decodes the current points with subgrid accuracy. Updates all the quantities for rotation
/// @var    probGrids        Probability grids filled with the dot-offset probabilities
/// @var    pictureCenter    Center of the image(the position to be decoded)
/// @var    rotDec           Current rotationDecoder
/// @var    means            Grid's main directions in picture coordinates*subdivision
/// @var    dotInfo          Information about each dot and the grid origin on the image
/// @var    divisorForResult Output will be scaled by this factor
/// @return The position of pictureCenter in the global coordinates of the grid multiplied by divisor for result
IntPoint probGridsDecode(ProbabilityGrids* probGrids, IntPoint pictureCenter, RotationDecoder* rotDec, IntPoint* means, DotInformation* dotInfo, int divisorForResult, IntPoint* sector);

/// @brief  Finds the Section of the pattern (Shift of the pattern at 0,0)
/// @var    topLeftPosition decoded position of a subgrid
/// @var    topMainShift    shift of the top row in the main number sequence
/// @var    leftMainShift   shift of the left collumn in the main number sequence
/// @return The 2 sections coordinates as a IntPoint
IntPoint findSection(IntPoint topLeftPosition, unsigned char topMainShift, unsigned char leftMainShift);

/// @brief  Assigns probabilities to offsetLists
/// @var    offsetsx    Pointer to list of offsets in x-direction
/// @var    offsetsy    Pointer to list of offsets in y-direction
/// @var    n           length of the two lists
IntPoint* probabilities(int* offsetsx, int* offsetsy, int n);

#ifdef __cplusplus
}
#endif

#endif // DECODER_H

