/**
 * @file localization.c
 * @brief Source of the localization functionality
 * @author Lukas Hostettler
 * @version 1.0
 * @date 2015-05-02
 */

#include <math.h>
#include <stdlib.h>

#include "compiler.h"
#include "config.h"
#include "dots_localization.h"

DotsPositionInfo* dotsPosInfoInit(){
    DotsPositionInfo* ans = (DotsPositionInfo*)malloc(sizeof(DotsPositionInfo)); //TODO: GET RID OF MALLOC
    if(!ans) //malloc trap
        return ans;

    initMeans(ans->means, MEAN_INIT_LENGTH);
    rotationDecoderReset(&ans->rotDec);
    ans->positionCertainty.x = 0;
    ans->positionCertainty.y = 0;
    ans->decoded = 0;
    ans->decodingError = DECODING_NO_ERROR;
    return ans;
}

DotsPositionInfo* dotsLocalize(DotsPositionInfo* previousInfo, unsigned char* image, int rows, int cols){
    if(!image || rows < 1 || cols < 1)
        return 0;
    if(!previousInfo)
        previousInfo = dotsPosInfoInit();

    previousInfo->decodingError = DECODING_NO_ERROR;
    previousInfo->decoded = 0;

    /*
     * Image processing
     */

    //Blob detection
    int subdivision = SUBDIVISION;
    int thresholdValue = imageMean(image, rows, cols) - THRESHOLD_OFFSET;
    ImgSegmentList segList = segmentImage(image, rows, cols, thresholdValue);

    previousInfo->positionCertainty.y = segList.numElements;
    if(segList.numElements < 64){
        previousInfo->decodingError = DECODING_NOT_ENOUGH_POINTS;
        return previousInfo;
    }
    if(segList.numElements >= MAX_NUM_POINTS){
        previousInfo->decodingError = DECODING_TOO_MANY_POINTS;
        return previousInfo;
    }

    IntPoint* pointArray = sortPointArray(segList, subdivision);
    IntPoint imgCenter = {rows/2, cols/2};

    //Distortion correction
    undistortPoints(pointArray, segList.numElements, rows, cols, CAMERA_DISTORTION_PARAMETER, subdivision);

    /*
     * Grid estimation
     */

    //Find initial grid
    int pointArrayNumElements = segList.numElements;
    Neighbours* neigh = establishNeighbourhood(pointArray, pointArrayNumElements);
    int numEdges;
    Edge* edges = extractGoodEdges(neigh, pointArray, pointArrayNumElements, &numEdges);
    linkedKMeans(edges, previousInfo->means, numEdges);

    //Refine grid estimate
    IntPoint cross = pointArray[pointArrayNumElements/2];
    findCross(neigh, pointArray, pointArrayNumElements, &cross);
    if(correctMeanLength(pointArray, previousInfo->means, 10, pointArrayNumElements, &cross)){

        /*
         * Probabilistic treatment
         */

        //Assign probabilities
        DotInformation dotInfo = dotInfoInit(pointArray, previousInfo->means, 10, pointArrayNumElements, cross);
        if(dotInfo.gridMaxCols < MAX_GRID_LENGTH && dotInfo.gridMaxRows < MAX_GRID_LENGTH){
            ProbabilityGrids probGrids = makeProbabilityGrids(dotInfo);
            previousInfo->positionCertainty.x = (int)probGrids.maxProb.numCols;

            /*
             * Decoding & probabilistic treatment continued
             */

            if(probGrids.maxProb.numCols >= 8 && probGrids.maxProb.numRows >= 8){
                IntPoint pictureCenter = {(rows*subdivision/2), (cols*subdivision/2)};

                //Find best area to decode (part of probabilistic treatment)
                //Correct orientation
                //Decode
                //Do subgrid resolution
                IntPoint decodedPos = probGridsDecode(&probGrids, pictureCenter, &previousInfo->rotDec, previousInfo->means, &dotInfo, FIXED_RESULT_DIVISOR, &previousInfo->sector);

                if(decodedPos.x >= 0)
                    previousInfo->position.x = decodedPos.x;
                else
                    previousInfo->decodingError = DECODING_X_FAILED;

                if(decodedPos.y >= 0)
                    previousInfo->position.y = decodedPos.y;
                else
                    previousInfo->decodingError = DECODING_Y_FAILED;

                if(decodedPos.x < 0 && decodedPos.y < 0)
                    previousInfo->decodingError = DECODING_BOTH_FAILED;

                previousInfo->decoded = 1;
            }
            else{
                previousInfo->decodingError = DECODING_WRONG_GRID_SIZE;
            }
            probabilityGridsFree(&probGrids);
        }
        else{
            previousInfo->decodingError = DECODING_WRONG_GRID_SIZE;
        }
    }
    else{
        previousInfo->decodingError = DECODING_COLLINEAR_GRID;
        rotationDecoderDiminish(&previousInfo->rotDec);
    }
    return previousInfo;
}

void dotsRelease(DotsPositionInfo** info){
    if(info && *info){
        free(*info);
        *info=0;
    }
}

float dotsGetFloatX(DotsPositionInfo* info){
    if(info)
        return info->position.x/(float)info->divisor;
    return -1;
}

float dotsGetFloatY(DotsPositionInfo* info){
    if(info)
        return info->position.y/(float)info->divisor;
    return -1;
}

int dotsGetIntX(DotsPositionInfo* info){
    if(info)
        return info->position.x;
    return -1;
}

int dotsGetIntY(DotsPositionInfo* info){
    if(info)
        return info->position.y;
    return -1;
}

int dotsGetDivisor(DotsPositionInfo* info){
    if(info)
        return info->divisor;
    return -1;
}

void dotsSetDivisor(DotsPositionInfo* info, int divisor){
    if(info && divisor)
        info->divisor = divisor;
}

float dotsGetAngle(DotsPositionInfo* info){
    if(info)
        return atan2(info->means[0].y, info->means[0].x);
    return 0;
}

float dotsGetCertaintyX(DotsPositionInfo* info){
    if(info)
        return info->positionCertainty.x;
    return 0;
}

float dotsGetCertaintyY(DotsPositionInfo* info){
    if(info)
        return info->positionCertainty.y;
    return 0;
}

float dotsGetCertaintyAngle(DotsPositionInfo *info){
    //to be changed;
    if(info)
        return info->rotDec.x+info->rotDec.y;
    return 0;
}

int dotsIsDecoded(DotsPositionInfo* info){
    if(info)
        return info->decoded;
    return 0;
}

DecodingError dotsGetError(DotsPositionInfo* info){
    if(info)
        return info->decodingError;
    return NO_INFO;
}

int dotsIsAboveGround(DotsPositionInfo* info){
    if(info){
        switch(info->decodingError){
            case DECODING_NOT_ENOUGH_POINTS:
            case DECODING_COLLINEAR_GRID:
                return 1;
            default:
                return 0;
        }
    }
    return 1;
}

int dotsGetSectorX(DotsPositionInfo* info){
    if(info){
        //TODO
    }
    return -1;
}

int dotsGetSectorY(DotsPositionInfo * info){
    if(info){
        //TODO
    }
    return -1;
}

