/**
 * @file grid.h
 * @brief Header responsible for the transmission between a point list and an ordered grid
 * @author Lukas Hostettler
 * @version 1.0
 * @date 2015-05-20
 */

#ifndef GRID_H
#define GRID_H

#ifdef __cplusplus
extern "C" {
#endif

#include "list.h"
#include "neighbours.h"
#include "intgrid.h"

/// @brief Corrects the length of the means
///
/// Given estimated means this function corrects their length.
/// It also provides the point of origin close to the center of points and aligned with the grid.
///
/// @var pointArray     Pointer to a list of the underlying PointCoordinates
/// @var means          Pointer to an array of two InPoints. Is the initial guess of the grid main-vectors (changed in the process)
/// @var power          Number of bit-shifts to use for the projection values (the higher power, the more likely to have overflow, but the more precise)
/// @var n              number of points in pointArray
/// @var origin         Pointer to an IntPoint holding the coordinates of a initial guess of the origin( the pointed value will change)
/// @returns            Success of the operation
int correctMeanLength(IntPoint* pointArray, IntPoint* means, int power, int n, IntPoint* origin);

///@brief Stores the information about dots in lists
typedef struct _DotInformation{
    int numElements;            //! Number of dots
    int gridMaxRows;            //! Number of rows
    int gridMaxCols;            //! Number of cols
    IntPoint gridOrigin;        //! Pixel-coordinates of the origin of the grid
    int* xoffsets;              //! Pointer to an array of x-offset positions
    int* yoffsets;              //! Pointer to an array of y-offset positions
    IntPoint* gridCoordinate;   //! Pointer to an array of grid coordinates
    //Prob?
} DotInformation;

/// @brief  Initializes a structure holding information on the dots
/// @var    pointArray  Pointer to an array of the underlying PointCoordinates
/// @var    means       Pointer to an array of two InPoints. Is the initial guess of the grid main-vectors.
/// @var    power       Bit-shift of the representation in grid-coordinates
/// @var    n           Length of the pointArray
/// @var    origin      Initial guess of the origin in pixel coordinates
DotInformation dotInfoInit(IntPoint* pointArray, IntPoint* means, int power, int n, IntPoint origin);

//void dotInfoFree(DotInformation * d);

///@brief Holds all probability grids needed for decoding
typedef struct _ProbabilityGrids{
    IntGrid maxProb; //! The absolute value of the product between prob1 and prob2 (tells the quality of the point)
    IntGrid prob1;   //! Probabilities of value 1 (usually represents X, but if the grid is read the wrong way, it can be others)
    IntGrid prob2;   //! Probabilities of value 2 (usually represents Y, but if the grid is read the wrong way, it can be others)
} ProbabilityGrids;

/// @brief Turns the probability grids according to the number of turns and sets the probablities according to reading direction
/// @var    grids           Pointer to a ProbabilityGrids structure (value might change)
/// @var    numRightTurns   Number of right turns to be made to the grids structure.
void probabilityGridsTurn(ProbabilityGrids* grids, int numRightTurns);

///@brief   Constructs probability grids from the dotInfo
///@var     dotInfo Structure holding all informations related to the dots coordinates and offsets
///@returns A ProbabilityGrids structure filled with all information
ProbabilityGrids makeProbabilityGrids(DotInformation dotInfo);

//TODO: GET RID OF THIS
///@brief Releases all ressources inside grid and sets the pointers to NULL
///@var grids Pointer to a ProbabilityGrids structure.(will be reset)
void probabilityGridsFree(ProbabilityGrids* grid);

#ifdef __cplusplus
}
#endif

#endif // GRID_H

