/**
 * @file imgproc.c
 * @brief Source of the process to extract the maxima of an image
 * @author Lukas Hostettler
 * @version 1.0
 * @date 2015-05-20
 */

#include <stdlib.h>

#include "config.h"
#include "point.h"
#include "imgproc.h"

#define NOSEGMENT MAX_NUM_IMAGE_SEGMENTS

///@var pointArray Static array containing the center of the detected dots in (pixel coordinates in fixed comma representation)
static IntPoint pointArray[MAX_NUM_POINTS];

///@var segmentList    Static array containing the detected image segments
static ImgSegment segmentList[MAX_NUM_IMAGE_SEGMENTS];

///@var equivalenceList Static array containg the indices of equivalence. if the value is always smaller or equal to the index.
static unsigned short equivalenceList[MAX_NUM_IMAGE_SEGMENTS];

void undistortPoints(IntPoint* pointList, int numPoint, unsigned short nRows, unsigned short nCols, float k, int subdivision){
    if(k == 0)
        return;

    int i;
    int rSquared;
    int largestRSquared = (nCols*nCols/4 + nRows*nRows/4)/subdivision;

    //Calculate 1/K in r^2 scale
    int realK = (int)(largestRSquared/k);

    IntPoint center = {nCols/2*subdivision, nRows/2*subdivision};
    IntPoint diff;
    for(i=0; i<numPoint; ++i){
        diff = intPointDiff(pointList[i], center);

        rSquared = (diff.x*diff.x + diff.y*diff.y)/subdivision/subdivision/subdivision;

        pointList[i].x = diff.x*realK/(rSquared + realK) + center.x;
        pointList[i].y = diff.y*realK/(rSquared + realK) + center.y;
    }
}

ImgSegmentList segmentImage(unsigned char* imgPointer, unsigned short nRows, unsigned short nCols, unsigned char thresholdValue){
    unsigned short i, j;
    unsigned short numSegments = 0;
    unsigned char* p;

    //init oldline
    unsigned short line1[nCols];
    unsigned short line2[nCols];

    unsigned short* oldline = line1;
    unsigned short* newline = line2;
    unsigned short* tmpline = 0;

    for (j=0; j<nCols; ++j){
        oldline[j] = NOSEGMENT;
        newline[j] = NOSEGMENT;
    }

    for(i = 0; i<nRows && numSegments<MAX_NUM_IMAGE_SEGMENTS; ++i){
        p = &imgPointer[i*nCols]; //point to next row

        //Special case for first pixel of row
        if(p[0] < thresholdValue){
            unsigned short up = oldline[0];
            if(up != NOSEGMENT){
                newline[0] = up;
                imgSegmentAddPixel(&segmentList[up], 0, i);
            }
            else{
                segmentList[numSegments] = imgSegmentCreate(0,i,1);
                newline[0] = numSegments;
                equivalenceList[numSegments] = numSegments;
                numSegments++;
                if(numSegments == MAX_NUM_IMAGE_SEGMENTS)
                    break;
            }
        }
        else
            newline[0]=NOSEGMENT;

        //case for the rest of the pixels
        for (j=1; j<nCols; ++j){
            if(p[j] < thresholdValue){
                unsigned short last = newline[j-1];
                unsigned short up = oldline[j];
                if(up != NOSEGMENT){
                    newline[j] = up;
                    imgSegmentAddPixel(&segmentList[up], j, i);
                    if(last != NOSEGMENT && last != up){

                        //let the smaller survive
                        unsigned short greater = last > up ? last : up; //TODO: REPLACE THIS WITH IF ELSE
                        unsigned short smaller = last < up ? last : up;
                        newline[j] = smaller;
                        equivalenceList[greater] = smaller;
                    }
                }
                else if(last != NOSEGMENT){
                    newline[j] = last;
                    imgSegmentAddPixel(&segmentList[last], j, i);
                }
                else{
                    segmentList[numSegments] = imgSegmentCreate(j, i, 1);
                    newline[j] = numSegments;
                    equivalenceList[numSegments] = numSegments;
                    numSegments++;
                    if(numSegments >= MAX_NUM_IMAGE_SEGMENTS)
                        break;
                }
            }
            else
                newline[j] = NOSEGMENT;
        }
        tmpline = oldline;
        oldline = newline;
        newline = tmpline;
        tmpline = 0;

    }

    //Equivalence Solving
    //As the smallest always survives equivalences so we need to go through the list in inverse order
    unsigned short numGoodSegments = 0;
    for(i = numSegments - 1; i != 0xFFFF; --i){ //TODO: THIS AND BELOW IS PRETTY UNREADABLE
        if(equivalenceList[i] == i){ //reached a base equivalence

            //save in the last unsaved position
            if(segmentList[i].n >= MIN_PIXELS_PER_SEGMENT)
                segmentList[numSegments - (++numGoodSegments)] = segmentList[i];

        }
        else{ //add to lower value segment
            imgSegmentAdd(&segmentList[equivalenceList[i]], segmentList[i]);
        }
    }
    ImgSegmentList ans;
    ans.list = &segmentList[numSegments - numGoodSegments];
    ans.numElements = numGoodSegments;
    return ans;
}

IntPoint* sortPointArray(ImgSegmentList list, int multiplier){
    int i, j;
    int num = list.numElements;
    if(list.numElements > MAX_NUM_POINTS)
        return 0;

    //sort to list using insertion sort (almost order(n) )
    for(i=0; i<num; ++i){
        j = i;
        IntPoint p;
        p.x = (list.list[i].x*multiplier)/list.list[i].n;
        p.y = (list.list[j].y*multiplier)/list.list[i].n;
        pointArray[i] =  p;
        while(j > 0 && pointArray[j].y < pointArray[j - 1].y){
            IntPoint temp = pointArray[j];
            pointArray[j] = pointArray[j - 1];
            pointArray[j - 1] = temp;
            --j;
        }
    }
    return pointArray;
}

unsigned char imageMean(unsigned char* image, int rows, int cols){
    //acceleration
    //return 0xBC;
    int i;
    unsigned int buffer = 0;
    unsigned int numpixel = rows*cols;

    //TODO: THIS IS A BIG BOTTLENECK: USE image AS INCREMENTED POINTER AND GET RID OF i
    for(i=0; i<numpixel; ++i){
        buffer += image[i];
    }

    return (unsigned char)(buffer/numpixel);
}

