/**
 * @file imgproc.h
 * @brief Header of the process to extract the maxima of an image
 * @author Lukas Hostettler
 * @version 1.0
 * @date 2015-05-20
 */

#ifndef IMGPROC_H
#define IMGPROC_H

#include "point.h"
#include "imgsegment.h"

#ifdef __cplusplus
extern "C" {
#endif

///@brief Structure holding a list of image segments
typedef struct {
    unsigned short numElements; //! Number of elements in the list
    ImgSegment* list;           //! Pointer to start of list
} ImgSegmentList;

/// @brief Calculates the mean of the image
/// @var    image           Pointer to an 8-bit image
/// @var    rows            Number of rows
/// @var    cols            Number of columns
/// @return Mean grayscale pixel value of the image
unsigned char imageMean(unsigned char* image, int rows, int cols);

/// @brief Finds 4-connected components of the image, gives at most MAX_NUM_SEGMENTS segments back
/// @var    imgPointer      Pointer to an 8-bit image
/// @var    nRows           Number of rows of pixels of the image
/// @var    nCols           Number of columns of pixels of the image
/// @var    thresholdValue  Deciding value for foreground or background
/// @return An ImgSegmentList structure containing the center of all 4-connected foreground elements
ImgSegmentList segmentImage(unsigned char* imgPointer, unsigned short nRows, unsigned short nCols, unsigned char thresholdValue);

/// @brief Transforms every point in the pointlist according to its distance to the center
/// @var    pointList       Pointer to an array of Intpoints (array is modified)
/// @var    numPoints       Number of entries in pointList
/// @var    nRows           Number of rows in image
/// @var    nCols           Number of columns in image
/// @var    k               Distortion factor of the second degree distortion
/// @var    subdivision     Factor used for the fixed point representation
void undistortPoints(IntPoint* pointList, int numPoint, unsigned short nRows, unsigned short nCols, float k, int subdivision);

/// @brief Sorts the image list according to the y-values
/// @var    list            Structure holding the information about the centers and width of the dots
/// @var    multiplier      Multiplier used for fixed point representation. The output-coordinates are stretched by this factor
/// @returns An array of horizontally orederd intpoints, the list is of length list.numsegments
IntPoint* sortPointArray(ImgSegmentList list, int multiplier);

#ifdef __cplusplus
}
#endif

#endif // IMGPROC_H
