/**
 * @file neighbours.c
 * @brief Source of neighbour and edge finding
 * @author Lukas Hostettler
 * @version 1.0
 * @date 2015-05-20
 */

#include "compiler.h"
#include "config.h"
#include "point.h"

#include "neighbours.h"

#define NO_DIRECTION (-1)

static Neighbours neighbourArray[MAX_NUM_POINTS];   ///< Holds points neighbour indices
static Edge edgeArray[MAX_NUM_POINTS*2];            ///< Holds the links between points

Neighbours* establishNeighbourhood(IntPoint* pointArray, int num){

    //This function jumps back and forth in the pointArray to find closest neighbors it stops as soon as the vertical distances make it impossible to find closer points
    int i, j, k, maxIndex = 0;
    int forward; //index pointer bigger than currently pointed position
    int backward;//index pointer smaller than currently pointed position
    int dx, dy;
    int64_t squareDists[4], maxSquareDist;
    Neighbours* neigb = neighbourArray;


    for(i=0;i<num;++i){ //For every index i in the pointList

        //init distances to intmax
        for(j=0;j<4;++j){
            squareDists[j] = INT64_MAX;
        }
        maxSquareDist = INT64_MAX;

        forward = i + 1;    //index of the point further in the pointList
        backward = i - 1;   //index of a point before in the pointlist

        while(forward < num || backward >= 0){
            if(forward < num){
                dx = pointArray[i].x - pointArray[forward].x;
                dy = pointArray[i].y - pointArray[forward].y;
                int64_t sqdist = dx*dx + dy*dy;
                if(sqdist < maxSquareDist){ //a closer neighbor than the ones in the neigbors list is found
                    squareDists[maxIndex] = sqdist;
                    neigb[i].neigbourIdx[maxIndex] = forward;

                    //find the farthest neighbor, for the next comparisions
                    maxIndex = 0;
                    maxSquareDist = squareDists[0];
                    for(k=1;k<4;++k){
                        if(maxSquareDist < squareDists[k]){
                            maxIndex = k;
                            maxSquareDist = squareDists[k];
                        }
                    }
                }
                ++forward; //point to next point
                if(dy*dy >= maxSquareDist) //stop forward search, if the y-distance is bigger than the farthest neighbor's distance
                    forward = num;
            }
            if(backward >= 0){
                dx = pointArray[i].x - pointArray[backward].x;
                dy = pointArray[i].y - pointArray[backward].y;
                int64_t sqdist = dx*dx + dy*dy;
                if(sqdist < maxSquareDist){ //a closer neighbor than the ones in the neigbors list is found
                    squareDists[maxIndex] = sqdist;
                    neigb[i].neigbourIdx[maxIndex] = backward;

                    //find the farthest neighbor, for the next comparisions
                    maxSquareDist = squareDists[0];
                    maxIndex = 0;
                    for(k=1;k<4;++k){
                        if(maxSquareDist < squareDists[k]){
                            maxIndex = k;
                            maxSquareDist = squareDists[k];
                        }
                    }
                }
                --backward; //point to previous point
                if(dy*dy >= maxSquareDist) //stop backward search, if the y-distance is bigger than the farthest neighbor's distance
                    backward = -1;
            }
        }
    }

    return neigb; //return pointer to neigborhood array
}

Edge* extractGoodEdges(const Neighbours* neigh, const IntPoint* coords, int num, int* numEdges){
    int i, j, k, idx;
    int numEdge = 0;
    Neighbours tmp;
    Edge* edges = edgeArray;
    for(i=0;i<num;++i){ //for each point
        for(j=0;j<4;++j){ //for each neighbour of point
            idx = neigh[i].neigbourIdx[j];
            if(idx > i){
                tmp = neigh[idx];
                for(k=0;k<4;++k){ //for all neighbours of neighbour
                    if(tmp.neigbourIdx[k] == i){
                        edges[numEdge].direction = NO_DIRECTION;
                        edges[numEdge].dx = coords[idx].x - coords[i].x;
                        edges[numEdge].dy = coords[idx].y - coords[i].y;
                        edges[numEdge].start = i;
                        edges[numEdge++].end = idx;

                        break;
                    }
                }
            }
        }
    }
    *numEdges = numEdge;
    return edges;
}

//TODO: change initialIdx to unsigned short
/// @brief checks if the points are in a valid 3x3 grid
/// @var    initialIndex    index of a point in the neigh list
/// @var    neigh           pointer to array of neighbor links
/// @return true if the 3x3 neighborhood-links around the initial point are correctly interconnected
int findCrossInnerLoop(int initialIdx, Neighbours* neigh){
    int i, j, secondGenCandidate, k = -1;
    unsigned short* firstGen = neigh[initialIdx].neigbourIdx; //list of the four direct neighbors of initial index
    unsigned short* secondGenPointer; //pointer to the
    unsigned short secondGen[13]; //list of the neighbors neigbors. Has an additional space in order to avoid segmentation faults
    for(i=0;i<4;++i){
        int foundFirst = 0;
        secondGenPointer = neigh[firstGen[i]].neigbourIdx;
        for(j=0;j<4;++j){
            secondGenCandidate = secondGenPointer[j];
            if(secondGenCandidate == initialIdx)
                foundFirst = 1;
            else
                secondGen[++k] = secondGenCandidate; //TODO: EXPAND ONE LINER
        }
        if(!foundFirst) //invalid as there is no returning to initial point
            return 0;
    }

    //check that 3x3 corner pieces are mentioned twice as neighbors of the direct neighbor
    unsigned short doubleCounter = 0;
    for(i=0;i<12;++i){
        for(j=i+1;j<12;++j)
            doubleCounter += secondGen[i] == secondGen[j];
    }

    return doubleCounter == 4;
}

int findCross(Neighbours* neigh, IntPoint* coords, int numNeigh, IntPoint* crossCoordinates){
    crossCoordinates->x = coords[numNeigh/2].x;
    crossCoordinates->y = coords[numNeigh/2].y;
    //return 1;
    int j, i = numNeigh/2, zigZag = 1;
    while(i < numNeigh && i >= 0){
        if(findCrossInnerLoop(i, neigh)){
            crossCoordinates->x = 4*coords[i].x;
            crossCoordinates->y = 4*coords[i].y;

            for(j=0;j<4;j++){
                intPointAdd(crossCoordinates, coords[neigh[i].neigbourIdx[j]].x, coords[neigh[i].neigbourIdx[j]].y);
            }
            crossCoordinates->x /= 8;
            crossCoordinates->y /= 8;

            return 1;
        }
        i += (i&1) ? -zigZag : +zigZag;
        ++zigZag;
    }
    return 0;
}

void initMeans(IntPoint* means, int r){
    means[0].x = -r;
    means[0].y = 0;
    means[1].x = 0;
    means[1].y = -r;
}

//TODO: COMPLETE DOC
/// @brief   sorts the elements between left and right if they are bigger or smaller than the value at pivotIndex. Part of the quickselect algorithm(Rosetta code!)
/// @var    list    pointer to the list to partition
/// @var    left    index of the leftmost element  of the partition
/// @var    right   index of the rightmost element of the partition
/// @var    pivotIndex  index of a random point
int quickPartition(int* list, int left, int right, int pivotIndex){
    int storeIndex, tmp, i;
    int pivotValue = list[pivotIndex];
    list[pivotIndex] = list[right];
    list[right] = pivotValue;
    storeIndex = left;
    for(i=left;i<right;++i){
        if(list[i] < pivotValue){
            tmp = list[storeIndex];
            list[storeIndex] = list[i];
            list[i] = tmp;
            ++storeIndex;
        }
    }
    tmp = list[right];
    list[right] = list[storeIndex];
    list[storeIndex] = tmp;
    return storeIndex;
}

//TODO: COMPLETE DOC
/// Attention! This function changes the order of the content of list
/// @brief  finds the value of the n-th-biggest element in a list
/// @var    list    pointer to an array of integers (will be changed!)
/// @var    left    leftmost index in the array to consider
/// @var    right   righmost index of the array to consider
/// @var    n       index of the element of the sorted list to retrun.
/// @return value of the n-th entry of a sorted list.
int quickSelect(int* list, int left, int right, int n){
    if (left == right)
        return list[left];

    int pivotIndex = (left + right)/2;
    pivotIndex = quickPartition(list, left, right, pivotIndex);
    if(n == pivotIndex)
        return list[n];
    else if(n < pivotIndex)
        return quickSelect(list, left, pivotIndex - 1, n);
    else
        return quickSelect(list, pivotIndex + 1, right, n);
}

int intListMedian(IntList list){
    if(list.numElements)
        return quickSelect(list.list,0,list.numElements-1,list.numElements/2);
    else
        return 0;
}

///@brief Takes the median of the edges and puts them into *means
///@var edges   Array of Edges with at least num entries
///@var means   Array of 2 IntPoints, will contain the result
///@var num     number of Edges
void medianMeans(Edge* edges, IntPoint* means, int num){
    int dx, dy, dir, i;
    Edge* e;
    int squareLength[4];
    IntList accumulators[4];
    for(i=0;i<4;++i){
        squareLength[i] = means[i].x*means[i].x + means[i].y*means[i].y;
        accumulators[i] = intListCreate();
    }

    //sort edges into acumulators
    for(e=edges;e<edges+num;++e){
        dir = e->direction;
        if(dir < 2){
            dx = e->dx-means[dir].x;
            dy = e->dy-means[dir].y;
        }else{
            dx = e->dx + means[dir - 2].x;
            dy = e->dy + means[dir - 2].y;
        }
        if((dx*dx + dy*dy) < 0.25*squareLength[dir]){ //TODO: parameter to be determined
            intListPush(accumulators + (dir&1), (1 - (dir&2))*e->dx);
            intListPush(accumulators + (dir&1) + 2, (1 - (dir&2))*e->dy);
        }
    }
    for(i=0;i<2;++i){
        //TODO assure that non-zero...
        means[i].x = intListMedian(accumulators[i]);
        means[i].y = intListMedian(accumulators[i + 2]);
    }

    //free lists
    for(i=0;i<4;++i)
        intListFree(accumulators + i);
}

void linkedKMeans(Edge* edges, IntPoint* oldMean, int num){

    //Edges have a natural structure to be positive y.
    int i, j, it, minIdx, maxNumItt = 10;
    int minSqdist, tmpSqdist, dx, dy;
    int hasChanged = 1; //in order to go through loop.
    IntPoint oldMeanExtended[4];
    for(i=0;i<2;i++){ //TODO: POSSIBLY UNROLL THIS LOOP
        oldMeanExtended[i] = oldMean[i];
        oldMeanExtended[i + 2] = oldMean[i];
        intPointMul(oldMeanExtended + i + 2, -1);
    }
    IntPoint mean[4], tmpMean;
    int pointCounter[4];

    for(it = 0; it < maxNumItt && hasChanged; ++it){//while not stable solution
        for(j=0;j<4;j++){
            pointCounter[j] = 0;
            mean[j].x = 0;
            mean[j].y = 0;
        }
        for(i=0;i<num;++i){//for all vectors
            //classify point
            minIdx = 0;
            dx = edges[i].dx - oldMeanExtended[0].x;
            dy = edges[i].dy - oldMeanExtended[0].y;
            minSqdist = dx*dx + dy*dy;
            for(j=1;j<4;++j){ //for all clusters
                dx = edges[i].dx - oldMeanExtended[j].x;
                dy = edges[i].dy - oldMeanExtended[j].y;
                tmpSqdist = dx*dx + dy*dy;
                if(tmpSqdist < minSqdist){
                    minIdx = j;
                    minSqdist = tmpSqdist;
                }
            }

            //add classified Point to cluster
            mean[minIdx].x += edges[i].dx;
            mean[minIdx].y += edges[i].dy;
            ++pointCounter[minIdx];

            // save cluster info in edge
            edges[i].direction = minIdx;

            //could do the changeTest here?
            //+1 ItTeration if not moved,but control over numpoints reclassified
            //possibility to limit number of comp with clusters...
        }

        //move averages
        hasChanged = 0;
        for(j=0;j<2;++j){ //TODO: POSSIBLY UNROLL THIS LOOP
            int numPoint = pointCounter[j] + pointCounter[j + 2];
            if(numPoint){
                tmpMean.x = (mean[j].x - mean[j+2].x)/numPoint;
                tmpMean.y = (mean[j].y - mean[j+2].y)/numPoint;
            }
            else{

                //A divide by zero means the points are badly conditioned, or the means are off
                //Therefore we continue the algorithm with a reinit.

                tmpMean.x = (-1 + 2*j)*oldMean[1 - j].y;
                tmpMean.y = ( 1 - 2*j)*oldMean[1 - j].x;
            }
            if(tmpMean.x != oldMean[j].x || tmpMean.y != oldMean[j].y)
                hasChanged = 1;
            oldMeanExtended[j] = tmpMean;
        }
    }

    medianMeans(edges, oldMeanExtended, num);
    oldMean[0] = oldMeanExtended[0];
    oldMean[1] = oldMeanExtended[1];
}

