/**
 * @file neighbours.h
 * @brief Header of neighbour and edge finding
 * @author Lukas Hostettler
 * @version 1.0
 * @date 2015-05-20
 */

#ifndef NEIGHBOURS_H
#define NEIGHBOURS_H

#ifdef __cplusplus
extern "C" {
#endif

#include "list.h"
#include "imgproc.h"

/// @brief Structure with indices of the four closest neighbors
typedef struct _Neighbours{
    //TODO: Change to unsigned
    short neigbourIdx[4]; //! Array of indices to the four closest neighbors
    // isBorderPoint; //! boolean variable discribing if point is on the border of the image
} Neighbours;

//TODO: Change start and end to unsigned short
/// @brief Structure holding all necessary information for edges
typedef struct _Edge{
    int start;      //! Index of higher point
    int end;        //! Index of lower point
    int dx;         //! Length of the edge projected to the x-axis
    int dy;         //! Length of the edge projected to the y-axis
    //TODO: REMOVE
    int direction;  //! Indication of direction (unused)
} Edge;

/// @brief  Finds the closest centerpoint of a complete 3x3 grid
/// @var    neigh               Pointer to array of Neighbours in the same order as coords
/// @var    coords              Pointer to array of IntPoint coordinates, in the same order as neigh
/// @var    numNeigh            Length of neigh and coords arrays
/// @var    crossCoordinates    Pointer to an IntPointCoordinate, receives the coordinates of the middle of the good cross, closest to the center of the image
/// @return Success of finding a point with a complete neighborhood around
int findCross(Neighbours* neigh, IntPoint* coords, int numNeigh, IntPoint* crossCoordinates);

/// @brief  Finds the four closest neighbours for num points in pointArray
/// @var    pointArray1         Vertically ordered array of point coordinates
/// @var    num                 Number of points contained in pointArray1
/// @return Pointer to the neighbour Array (attention: due to the static allocation there will only be one neighborhood array)
Neighbours* establishNeighbourhood(IntPoint* pointArray1, int num);

/// @brief  Checks neighbours if they are mutually neighbours, gives back a list of those connections
/// @var    neigh               Pointer to array of neighborhood connections, with indices pointing to coords
/// @var    coords              Pointer to point coordinates-array
/// @var    num                 Number of entries in the neigh and coords arrays
/// @var    numEdges            Number of edges contained in the returned value
/// @return An array of edges which are mutually neighbors (attention: due to static allocation there can be only one such array)
Edge* extractGoodEdges(const Neighbours* neigh, const IntPoint* coords, int num, int* numEdges);

///@brief   Initializes means to be 90° offset to each other and of lenth r
///@var     means               Pointer to two-element array representing the grid-directions
///@var     r                   Length to be given to the grid-directions
void initMeans(IntPoint* means, int r);

///@brief Overrides *oldMeans with a new estimate of the directions based on edges
///
/// Works pretty nicely, length of the vectors are slighly offset
///
///@var     edges               Pointer to start of array of edges
///@var     oldMean             Pointer to array of 2 intpoints containing an initial guess of the directions
///@var     num                 Number of edges contained in edges
void linkedKMeans(Edge* edges, IntPoint* oldMean, int num);

//TODO: Move to list.h
/// @brief Gets the median of list; lists ordering is modified in this function
/// @var    list   List of integer values
/// @return Median value of the provided intlist.
int intListMedian(IntList list);

#ifdef __cplusplus
}
#endif

#endif // NEIGHBOURS_H

