﻿/**
 * @file webcam_sample.c
 * @brief Source for the sample webcam image acquisition and calling libdots
 * @author Lukas Hostettler
 * @author Ayberk Özgür
 * @version 1.0
 * @date 2015-09-03
 */

#include <iostream>
#include <iomanip>
#include <math.h>

#include "webcam_sample.h"

#define IMG_SIZE 3
#define NOSEGMENT 0xFFF

using namespace std;

WebcamSample::WebcamSample(int camNo){
    cap = VideoCapture(camNo);
    if(!cap.isOpened()){
        cout << "Cannot open the video cam" << endl;
        return;
    }
    cap.set(CV_CAP_PROP_FRAME_HEIGHT, IMG_SIZE*120.0);
    cap.set(CV_CAP_PROP_FRAME_WIDTH, IMG_SIZE*160.0);

    double dWidth = cap.get(CV_CAP_PROP_FRAME_WIDTH); //get the width of frames of the video
    double dHeight = cap.get(CV_CAP_PROP_FRAME_HEIGHT); //get the height of frames of the video

    cout << "Frame size : " << dWidth << "x" << dHeight << endl;

    posInfo = 0;
}

WebcamSample::~WebcamSample(){
    cap.release();
}

void WebcamSample::drawRobot(int width, int height, int radius){
    static Scalar background(255,255,255);
    static Scalar black(0,0,0);

    Mat I(height, width, CV_8UC3, background);

    float robotX = (float)dotsGetIntX(posInfo)/FIXED_RESULT_DIVISOR*GRID_SIZE;
    float robotY = (float)dotsGetIntY(posInfo)/FIXED_RESULT_DIVISOR*GRID_SIZE;
    float robotTheta = dotsGetAngle(posInfo);

    if(robotX >= 0 && robotY >= 0){
        Point pos(robotX, robotY);
        Point dir(radius*cosf(robotTheta), radius*sinf(-robotTheta));
        circle(I, pos, radius, black);
        line(I, pos, pos + dir, black);
    }

    imshow("Robot on A4 Sheet", I);
}

bool WebcamSample::getRawFrame(Mat& frame){
    bool ans = false;
    if(cap.isOpened())
        ans = cap.read(frame);
    return ans;
}

bool WebcamSample::getRawFrame(Mat& frame, Rect roi){
    bool ans = false;
    Mat tmp;
    if(cap.isOpened()){
        ans = cap.read(tmp);
        frame = tmp(roi).clone();
    }
    return ans;
}

void drawPoints(Mat& I, ImgSegmentList segs, Scalar color = Scalar(0,0,0)){
    for(ImgSegment* i = segs.list; i < &segs.list[segs.numElements]; i++)
        circle(I, Point(i->x/i->n, i->y/i->n), 5, color);
}

void drawLines(Mat& I, IntPoint* pointArray, Neighbours* neigh, int n, int mul, Scalar color = Scalar(0,0,0)){
    for(int i=0;i<n;i++){
        Point p1(pointArray[i].x/mul, pointArray[i].y/mul);
        Neighbours nei = neigh[i];
        for(int j=0;j<4;j++){
            Point p2(pointArray[nei.neigbourIdx[j]].x/mul, pointArray[nei.neigbourIdx[j]].y/mul);
            line(I, p1, p2, color);
        }
    }
}

void drawLines(Mat& I, IntPoint* pointArray, Edge* edges, int n, int mul, Scalar color = Scalar(0,0,0)){
    for(Edge* e = edges; e < &edges[n]; e++){
        Point p1(pointArray[e->start].x/mul, pointArray[e->start].y/mul);
        Point p2(pointArray[e->end].x/mul, pointArray[e->end].y/mul);
        line(I, p1, p2, color);
    }
}

void drawDirections(Mat& I, IntPoint* pointArray, int n, int mul, IntPoint direction, Scalar color = Scalar(0,0,0)){
    Point dir(direction.x/mul, direction.y/mul);
    for(IntPoint* point = pointArray; point < &pointArray[n]; point++){
        Point p1(point->x/mul, point->y/mul);
        Point p2 = p1 + dir;
        line(I, p1, p2, color);
    }
}

Point transformFromIdx(int col, int row, IntPoint* means, IntPoint origin, int subdivision){
    int px = (origin.x + row*means[0].x + col*means[1].x)/subdivision;
    int py = (origin.y + row*means[0].y + col*means[1].y)/subdivision;
    return Point(px, py);
}

void printIntGrid(Mat& I, IntGrid g, IntPoint* means, IntPoint origin, int subdivision, int min = 0, int max = 1024*256){
    int row, col;
    for(row=0;row<g.numRows;++row){
        for(col=0;col<g.numCols;++col){
            Point center = transformFromIdx(col, row, means, origin, subdivision);
            int colorVal = (255*g.array[row][col])/(max - min);
            circle(I, center, 5, Scalar(0, colorVal, 255 - colorVal), -1);
        }
    }
}

void printClassification(Mat& I, ProbabilityGrids grids, IntPoint* means, IntPoint origin, int subdivision){
    int row, col;
    string text;
    int fontFace = CV_FONT_HERSHEY_PLAIN;
    double fontScale = 0.5;

    origin.x -= 300;
    origin.y += 300;

    for(row=0;row<grids.maxProb.numRows;++row){
        for(col=0;col<grids.maxProb.numCols;++col){
            Point center = transformFromIdx(col, row, means, origin, subdivision);
            if(     grids.prob1.array[row][col] < 0 && grids.prob2.array[row][col] < 0)
                text = "L";
            else if(grids.prob1.array[row][col] > 0 && grids.prob2.array[row][col] < 0)
                text = "D";
            else if(grids.prob1.array[row][col] < 0 && grids.prob2.array[row][col] > 0)
                text = "U";
            else if(grids.prob1.array[row][col] > 0 && grids.prob2.array[row][col] > 0)
                text = "R";
            putText(I, text, center, fontFace, fontScale, Scalar(0,0,0));
        }
    }
}

void printSquare(Mat& I, int row, int col, IntPoint* means, IntPoint origin, int subdivision, Scalar color = Scalar(255,0,0)){
    Point pt0 = transformFromIdx(col, row, means, origin, subdivision);
    Point pt1 = transformFromIdx(col + 7, row, means, origin, subdivision);
    Point pt2 = transformFromIdx(col + 7, row + 7, means, origin, subdivision);
    Point pt3 = transformFromIdx(col, row + 7, means, origin, subdivision);

    circle(I, pt0, 10, color, 2);
    line(I, pt0, pt1, color, 2);
    line(I, pt1, pt2, color, 2);
    line(I, pt2, pt3, color, 2);
    line(I, pt3, pt0, color, 2);
}

void WebcamSample::printInfo(){
    float robotX = ((float)dotsGetIntX(posInfo))/FIXED_RESULT_DIVISOR*GRID_SIZE;
    float robotY = ((float)dotsGetIntY(posInfo))/FIXED_RESULT_DIVISOR*GRID_SIZE;
    float robotTheta = dotsGetAngle(posInfo)/3.14159*180;

    cout << "x: " << robotX << " mm, y:" << robotY << " mm, theta: " << robotTheta << " degrees";

    switch(posInfo->decodingError){
        case DECODING_NO_ERROR:
            cout << endl;
            break;
        case DECODING_TOO_MANY_POINTS:
            cout << "; Decoding error: Too many points" << endl;
            break;
        case DECODING_NOT_ENOUGH_POINTS:
            cout << "; Decoding error: Not enough points" << endl;
            break;
        case DECODING_COLLINEAR_GRID:
            cout << "; Decoding error: Collinear grid" << endl;
            break;
        case DECODING_WRONG_GRID_SIZE:
            cout << "; Decoding error: Wrong grid size" << endl;
            break;
        case DECODING_X_FAILED:
            cout << "; Decoding error: Decoding X failed" << endl;
            break;
        case DECODING_Y_FAILED:
            cout << "; Decoding error: Decoding Y failed" << endl;
            break;
        case DECODING_BOTH_FAILED:
            cout << "; Decoding error: Decoding both failed" << endl;
            break;
        case NO_INFO:
            cout << "; Decoding error: No info" << endl;
            break;
    }
}


bool WebcamSample::run(){
    if(waitKey(10) != -1)
        return false;

    int i;
    int chan = 1; //Use the green channel, it has the best contrast in our case
    Mat frame;
    bool ans = getRawFrame(frame);
    vector<Mat> channels;
    split(frame, channels);
    Mat originalFrame = frame.clone();
    unsigned char* image = channels[chan].ptr();
    int rows = channels[chan].rows;
    int cols = channels[chan].cols;

    //
    //We do not use dotsLocalize() directly but replicate its implementation to illustrate the steps in the imshow
    //

    if(!posInfo)
        posInfo = dotsPosInfoInit();

    posInfo->decodingError = DECODING_NO_ERROR;
    posInfo->decoded = 0;

    /*
     * Image processing
     */

    //Blob detection
    int subdivision = SUBDIVISION;
    int thresholdValue = imageMean(image, rows, cols) - CUSTOM_THRESHOLD_OFFSET;
    ImgSegmentList segList = segmentImage(image, rows, cols, thresholdValue);

    posInfo->positionCertainty.y = segList.numElements;
    if(segList.numElements < 64){
        posInfo->decodingError = DECODING_NOT_ENOUGH_POINTS;
        imshow("Original Image (assuming mirrored)", originalFrame);
        printInfo();
        return true;
    }
    if(segList.numElements >= MAX_NUM_POINTS){
        posInfo->decodingError = DECODING_TOO_MANY_POINTS;
        imshow("Original Image (assuming mirrored)", originalFrame);
        printInfo();
        return true;
    }

    IntPoint* pointArray = sortPointArray(segList, subdivision);

    //Distortion correction
    undistortPoints(pointArray, segList.numElements, rows, cols, CUSTOM_CAMERA_DISTORTION_PARAMETER, subdivision);

    /*
     * Grid estimation
     */

    //Find initial grid
    int pointArrayNumElements = segList.numElements;
    Neighbours* neigh = establishNeighbourhood(pointArray, pointArrayNumElements);
    int numEdges;
    Edge* edges = extractGoodEdges(neigh, pointArray, pointArrayNumElements, &numEdges);
    linkedKMeans(edges, posInfo->means, numEdges);

    //Draw neighborhood lines
    drawLines(frame, pointArray, neigh, segList.numElements, subdivision);

    //Refine grid estimate
    IntPoint cross = pointArray[pointArrayNumElements/2];
    int findCrossSuccess = findCross(neigh, pointArray, pointArrayNumElements, &cross);
    if(findCrossSuccess)
        circle(frame, Point(cross.x/subdivision, cross.y/subdivision), 7, Scalar(0,0,255), 3);
    if(correctMeanLength(pointArray, posInfo->means, 10, pointArrayNumElements, &cross)){

        /*
         * Probabilistic treatment
         */

        //Assign probabilities
        DotInformation dotInfo = dotInfoInit(pointArray, posInfo->means, 10, pointArrayNumElements, cross);
        if(dotInfo.gridMaxCols < MAX_GRID_LENGTH && dotInfo.gridMaxRows < MAX_GRID_LENGTH){
            ProbabilityGrids probGrids = makeProbabilityGrids(dotInfo);

            //Print probabilities and assigned symbols
            printIntGrid(frame, probGrids.maxProb, posInfo->means, dotInfo.gridOrigin, subdivision);
            printClassification(frame, probGrids, posInfo->means, dotInfo.gridOrigin, subdivision);

            posInfo->positionCertainty.x = (int)probGrids.maxProb.numCols;

            if(probGrids.maxProb.numCols >= 8 && probGrids.maxProb.numRows >= 8){
                IntPoint pictureCenter = {(rows*subdivision/2), (cols*subdivision/2)};

                /*
                 * Decoding & probabilistic treatment continued
                 */

                //Find best area to decode (part of probabilistic treatment)
                //Correct orientation
                //Decode
                //Do subgrid resolution
                IntPoint decodedPos = probGridsDecode(&probGrids, pictureCenter, &posInfo->rotDec, posInfo->means, &dotInfo, FIXED_RESULT_DIVISOR, &posInfo->sector);

                //Find the best area to decode separately again just to draw on screen
                int nCol = 0, nRow = 0;
                intGridFindBestNxN(probGrids.maxProb, &nCol, &nRow, 8);
                printSquare(frame, nRow, nCol, posInfo->means, dotInfo.gridOrigin, subdivision);

                if(decodedPos.x >= 0)
                    posInfo->position.x = decodedPos.x;
                else
                    posInfo->decodingError = DECODING_X_FAILED;

                if(decodedPos.y >= 0)
                    posInfo->position.y = decodedPos.y;
                else
                    posInfo->decodingError = DECODING_Y_FAILED;

                if(decodedPos.x < 0 && decodedPos.y < 0)
                    posInfo->decodingError = DECODING_BOTH_FAILED;

                posInfo->decoded = 1;
            }
            else{
                posInfo->decodingError = DECODING_WRONG_GRID_SIZE;
            }
            probabilityGridsFree(&probGrids);
        }
        else{
            posInfo->decodingError = DECODING_WRONG_GRID_SIZE;
        }
    }
    else{
        posInfo->decodingError = DECODING_COLLINEAR_GRID;
        rotationDecoderDiminish(&posInfo->rotDec);
    }

    printInfo();
    drawRobot();
    imshow("Original Image (assuming mirrored)", originalFrame);
    imshow("Processed Image (press any key to exit)", frame);

    return true;
}

