/**
 * @file webcam_sample.h
 * @brief Header for the sample webcam image acquisition and calling libdots
 * @author Lukas Hostettler
 * @author Ayberk Özgür
 * @version 1.0
 * @date 2015-09-03
 */

#ifndef WEBCAM_SAMPLE_H
#define WEBCAM_SAMPLE_H

#include <opencv2/opencv.hpp>

#include "list.h"
#include "imgsegment.h"
#include "imgproc.h"
#include "neighbours.h"
#include "grid.h"
#include "decoder.h"
#include "config.h"
#include "dots_localization.h"

using namespace cv;

#define GRID_SIZE 0.508                         ///< Average distance in mm between two adjacent dots, in this example
#define CUSTOM_THRESHOLD_OFFSET 40              ///< This test webcam needs more offset
#define CUSTOM_CAMERA_DISTORTION_PARAMETER 0    ///< This test webcam does not need dist. correction

class WebcamSample {

public:

    /// @brief  Constructor, establishing connection to a webcam
    /// @var    camNo   number of the camera to connect
    /// @return a WebcamSample class
    WebcamSample(int camNo = 0);

    /// @brief  Destructor, closes connection to webcam. Frees the memory
    ~WebcamSample();

    /// @brief Runs this samples
    /// @return Whether to keep running, or a quit signal was received
    bool run();

private:

    /// @brief  grabs a frame from the camera and stores it in frame
    /// @var    frame   reference to a cv::Mat class
    /// @return success of operation
    bool getRawFrame(Mat& frame);

    /// @brief  grabs a frame from the camera and stores it in frame
    /// @var    frame   reference to a cv::Mat class
    /// @var    roi     Region of interest, only the region of interest will be treated
    /// @return success of operation
    bool getRawFrame(Mat& frame, Rect roi);

    /// @brief  opens a window if not yet opened and draws the current robot position
    /// @var    width   width of the coded paper in mm
    /// @var    heigh   height of the coded paper in mm
    /// @var    radius  radius of the device
    void drawRobot(int width = 210, int height = 297, int radius = 30);

    /// @brief Prints pose and error if present
    void printInfo();

    /// @brief  structure for the webcam handling
    cv::VideoCapture cap;

    /// @brief  pointer to structure containing all information relating to the decoding
    DotsPositionInfo* posInfo;
};

#endif // WEBCAM_SAMPLE_H

